/*-
 * Copyright (C)2008..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2008..2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: db_pgsql.c,v 1.84 2021/10/09 01:59:57 babolo Exp $"

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MIFE_COMPAT      4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1
#define PGOBLIN_MODULE   pgoblin_module

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <poll.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "pgoblin.h"
#include "level.h"

#define PRINTNOTNULL(A) (!(A) ? "" : (A))
#define PRINTANDNULL(A) PRINTNOTNULL(A), ((A) ? "~" : "")
#define PQSTATNDNULL(A) PRINTANDNULL(PQparameterStatus(pcnn->pgcn, (A)))
#define ALIGN8(A)  popt[i].A ? popt[i].A : ""                                                         \
      , ( (!popt[i].A || (strlen(popt[i].A) < 8))                                                     \
        ? "		"                                                                             \
        : (strlen(popt[i].A) < 16) ? "	" : " "                                                       \
        )

#define PQEXEC(RESULT, QUERY, CONDITION, CCODE, RCONN) do {                                           \
    ifBLIN_QX4("PQexec: %s\n", (char*)(QUERY));                                                       \
    (RESULT) = PQexec(RCONN, (QUERY));                                                                \
    if  (!(RESULT)) {                                                                                 \
        ifBLIN_QX0("result is absent - fatal error.\n");                                              \
        return(CCODE);                                                                                \
    }                                                                                                 \
    ifBLIN_QX4("PQresultStatus %d", PQresultStatus(RESULT));                                          \
    if  (PQresultStatus(RESULT) != (CONDITION)) {                                                     \
        if  (PQresultStatus(RESULT) == PGRES_FATAL_ERROR) {                                           \
            ifBLIN_QX0("PQexec failed: %s", PQresultErrorMessage(RESULT));                            \
            return(CCODE);                                                                            \
        } else if ((CONDITION) == PGRES_COMMAND_OK && PQresultStatus(RESULT) == PGRES_TUPLES_OK) {    \
            ifBLIN_QX1("PQexec wait for COMMAND_OK but received TUPLES_OK");                          \
        } else if ((CONDITION) == PGRES_COMMAND_OK && PQresultStatus(RESULT) == PGRES_EMPTY_QUERY) {  \
            ifBLIN_QX1("PQexec wait for COMMAND_OK but received EMPTY_QUERY");                        \
        } else ifBLIN_QO0 {                                                                           \
            warnx("PQexec wait for %d but received %d", (CONDITION), PQresultStatus(RESULT));         \
            return(CCODE);                                                                            \
}   }   } while(0);

typedef struct pgoblin_db_conn {
    BLIN_flag     flags;
/*                PGOBLIN_DB_TYPE                                                                    */
    int           numoi;
    PGconn       *pgcn;
    u_int32_t    *toid;
    char        **tname;
    void         *dummy1;
} pgoblin_db_conn;

struct pgoblin_mode {
    BLIN_flag        flags;
/*                   PGOBLIN_DB_TYPE                                                                 */
    PGresult        *res;
    char            *mess;
    pgoblin_db_conn *pcnn;
};

pgoblin_dbases PGOBLIN_MODULE;

static pgoblin_mode *
/**********************************************************************
 **                                                                  **/
resv(pgoblin_conn *db) {                                            /**
 **                                                                  **
 **********************************************************************/
    pgoblin_mode    *rsv  = NULL;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    if  (!(rsv = calloc(1, sizeof(pgoblin_mode)))) {
        ifBLIN_QW0("");
    } else {
        rsv->flags = db->flags & (BLIN_MASK | PGOBLIN_DB_TYPE);
        rsv->pcnn = db->conn;
    }
    return(rsv);
#   undef blin_internal_flags
}

static void
/**********************************************************************
 **                                                                  **/
NoticeReceiver(void *db, const PGresult *res) {                     /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (((pgoblin_conn *)db)->flags & BLIN_MASK)
    int severity; /*  ,   */

    severity = babolo_testword(&level, PQresultErrorField(res, PG_DIAG_SEVERITY_NONLOCALIZED));
    ifBLIN_QO2 {
        severity = -1;
    } else ifBLIN_QO1 {
        severity -= LOG_WARNING;
    } else ifBLIN_QO0 {
        severity -= LOG_ERR;
    }
    if  (0 > severity) {
        ifBLIN_QX0( "%s %s %s\n%s"
                  , PQresultErrorField(res, PG_DIAG_SEVERITY)
                  , PQresultErrorField(res, PG_DIAG_SQLSTATE)
                  , PQresultErrorField(res, PG_DIAG_MESSAGE_PRIMARY)
                  , PQresultErrorField(res, PG_DIAG_MESSAGE_DETAIL)
                  );
        ifBLIN_QX1( "%s\n%s\n"
                  , PQresultErrorField(res, PG_DIAG_MESSAGE_HINT)
                  , PQresultErrorField(res, PG_DIAG_STATEMENT_POSITION)
                  );
#ifdef PG_DIAG_INTERNAL_POSITION
        ifBLIN_QX1( "%s\n%s"
                  , PQresultErrorField(res, PG_DIAG_INTERNAL_QUERY)
                  , PQresultErrorField(res, PG_DIAG_INTERNAL_POSITION)
                  );
#endif
#ifdef PG_DIAG_SCHEMA_NAME
        ifBLIN_QX2( "shema=%s table=%s column=%s type=%s constraint=%s"
                  , PQresultErrorField(res, PG_DIAG_SCHEMA_NAME)
                  , PQresultErrorField(res, PG_DIAG_TABLE_NAME)
                  , PQresultErrorField(res, PG_DIAG_COLUMN_NAME)
                  , PQresultErrorField(res, PG_DIAG_DATATYPE_NAME)
                  , PQresultErrorField(res, PG_DIAG_CONSTRAINT_NAME)
                  );
#endif
        ifBLIN_QX2( "%s:%s:%s: %s"
                  , PQresultErrorField(res, PG_DIAG_SOURCE_FILE)
                  , PQresultErrorField(res, PG_DIAG_SOURCE_LINE)
                  , PQresultErrorField(res, PG_DIAG_SOURCE_FUNCTION)
                  , PQresultErrorField(res, PG_DIAG_CONTEXT)
                  );
    }
#   undef blin_internal_flags
}

#define U0 0
#define U1 1
#define U2 2
#define U3 3
#define U4 4

static u_char miss[][PGOBLIN_CR_MAX >> PGOBLIN_CR_SHIFT] =
/*  OK   TUPL OUT  IN   STR  BIN  BTUPL */
{{  U1,  U1,  U1,  U1,  U1,  U1,  U1 } /* PGRES_EMPTY_QUERY    */
,{  U0,  U2,  U2,  U2,  U2,  U2,  U2 } /* PGRES_COMMAND_OK     */
,{  U1,  U0,  U2,  U2,  U0,  U0,  U0 } /* PGRES_TUPLES_OK      */
,{  U2,  U2,  U0,  U2,  U2,  U2,  U2 } /* PGRES_COPY_OUT       */
,{  U2,  U2,  U2,  U0,  U2,  U2,  U2 } /* PGRES_COPY_IN        */
,{  U3,  U3,  U3,  U3,  U3,  U3,  U3 } /* PGRES_BAD_RESPONSE   */
,{  U3,  U3,  U3,  U3,  U3,  U3,  U3 } /* PGRES_NONFATAL_ERROR */
,{  U4,  U4,  U4,  U4,  U4,  U4,  U4 } /* PGRES_FATAL_ERROR    */
};

static ExecStatusType es[] =
{ PGRES_COMMAND_OK, PGRES_TUPLES_OK, PGRES_COPY_OUT , PGRES_COPY_IN
, PGRES_TUPLES_OK , PGRES_TUPLES_OK, PGRES_TUPLES_OK
};

static int
/*****************************************************************************
 **                                                                         **/
extatus( u_int32_t   flags                                                 /**/
       , PGresult   *res                                                   /**/
       , const char *who                                                   /**/
       , BLIN_flag   cond                                                  /**/
       , const char *func                                                  /**/
       , int         line                                                  /**/
       ) {                                                                 /**
 **                                                                         **
 *****************************************************************************/
    int            ex = 0;
    ExecStatusType st;
    u_char         cn = (cond & PGOBLIN_CR_MASK) >> PGOBLIN_CR_SHIFT;

#   define blin_internal_flags (flags & BLIN_MASK)
    st = PQresultStatus(res);
    ifBLIN_QX4("PQresultStatus %d %s, diff %d", st, PQresStatus(st), miss[st][cn]);
    switch (miss[st][cn]) {
    case U1:
        ifBLIN_QX1( "%s:%d:%s waits for %d %s but received %d %s"
                  , func
                  , line
                  , who
                  , es[cn]
                  , PQresStatus(es[cn])
                  , st
                  , PQresStatus(st)
                  );
        break;
    case U2:
        ifBLIN_QX0( "%s:%d:%s waits for %d %s but received %d %s"
                  , func
                  , line
                  , who
                  , es[cn]
                  , PQresStatus(es[cn])
                  , st
                  , PQresStatus(st)
                  );
        ex = miss[st][cn];
        break;
    case U4:
    case U3:
        ifBLIN_QX0( "%s:%d:%s failed: %s"
                  , func
                  , line
                  , who
                  , PQresultErrorMessage(res)
                  );
        /* erinfo(... ) XXXX */
        ex = miss[st][cn];
        break;
    }
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
init(BLIN_flag flags) {                                             /**
 **                                                                  **
 **********************************************************************/
    int er[] = { PGRES_EMPTY_QUERY   , PGRES_COMMAND_OK  , PGRES_TUPLES_OK     , PGRES_COPY_OUT
               , PGRES_COPY_IN       , PGRES_BAD_RESPONSE, PGRES_NONFATAL_ERROR, PGRES_FATAL_ERROR
               };
    int ex = EX_OK;
    int i;

#   define blin_internal_flags (flags & BLIN_MASK)
    for (i = 0; i < 8; ++i) {
        if  (er[i] != i) {
            ifBLIN_QX0("Invalid result code [%d] <> %d", er[i], i);
            ++ex;
    }   }
    if  (!ex) {
        PGOBLIN_MODULE.flags = (PGOBLIN_MODULE.flags & ~(BLIN_MASK | PGOBLIN_DB_TYPE))
                               | (flags & (BLIN_MASK | PGOBLIN_DB_TYPE))
        ;
    }
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
fini() {                                                            /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

static int
/**********************************************************************
 **                                                                  **/
shurecon(pgoblin_conn *db) {                                        /**
 **                                                                  **
 **********************************************************************/
    pgoblin_db_conn  *pcnn = db->conn;
    int               ex = EX_OK;
    PQconninfoOption *popt;
    PGresult         *res;
    int               i;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    if  (!!pcnn && !!pcnn->pgcn && (PQstatus(pcnn->pgcn) == CONNECTION_OK)) return(0);
    if  (!db->conn && !(pcnn = db->conn = calloc(1, sizeof(pgoblin_db_conn)))) {
        ifBLIN_QX0("no MEM");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ifBLIN_QO4 {
        fprintf( stderr
               , "Connection defaults:\n"
                 "size=chr keyword	envvar		compiled val		label\n"
               );
        popt = PQconndefaults();
        for (i = 0; popt[i].keyword; i++) {
            fprintf( stderr
                   , "%4d=%s~	%s%s%s%s%s%s%s%s%s\n"
                   , popt[i].dispsize
                   , popt[i].dispchar
                   , ALIGN8(keyword)
                   , ALIGN8(envvar)
                   , PRINTNOTNULL(popt[i].compiled)
                   , !(popt[i].compiled) ? "	" : strlen(popt[i].compiled) < 7 ? "~	" : "~ "
                   , PRINTNOTNULL(popt[i].val)
                   , !(popt[i].val)
                   ? "		"
                   : strlen(popt[i].val) < 7 ? "~		" : "~	"
                   , popt[i].label
                   );
        }
        PQconninfoFree(popt);
        fprintf( stderr
               , "PGREALM=%s%s PGDATESTYLE=%s%s PGTZ=%s%s PGCLIENTENCODING=%s%s PGGEQO=%s%s\n"
               , PRINTANDNULL(getenv("PGREALM"))
               , PRINTANDNULL(getenv("PGDATESTYLE"))
               , PRINTANDNULL(getenv("PGTZ"))
               , PRINTANDNULL(getenv("PGCLIENTENCODING"))
               , PRINTANDNULL(getenv("PGGEQO"))
               );
    }
    MARK_R_CONN_GO(db);;
    ;;  if  (db->flags & PGOBLIN_PARMIN && db->dbname) {
    ;;      pcnn->pgcn = PQconnectdb(db->dbname);
    ;;  } else {
    ;;      pcnn->pgcn = PQsetdbLogin(db->host, db->port, NULL, NULL, db->dbname, db->username, NULL);
    ;;  }
    ;;  db->flags = PGOBLIN_CLOSECONN
 /* ;; */         | (db->flags & ~(BLIN_MASK | PGOBLIN_DB_TYPE))
 /* ;; */         | (PGOBLIN_MODULE.flags & (BLIN_MASK | PGOBLIN_DB_TYPE))
    ;;  ;
    MARK_R_CONN_WENT(db);;
    ifBLIN_QO4 {
        fprintf(stderr, "PQoptions=%s~\n", PQoptions(pcnn->pgcn));
        fprintf( stderr
               , "PQdb=%s~ PQuser=%s~ PQpass=%s%s "
                 "PQhost=%s%s PQport=%s%s "
                 "PQtty=%s%s\n"
               , PQdb(pcnn->pgcn), PQuser(pcnn->pgcn)
               , PRINTANDNULL(PQpass(pcnn->pgcn))
               , PRINTANDNULL(PQhost(pcnn->pgcn))
               , PRINTANDNULL(PQport(pcnn->pgcn))
               , PRINTANDNULL(PQtty(pcnn->pgcn))
               );
        fprintf( stderr
               , "server_version=%s%s "
                 "client_encoding=%s%s "
                 "is_superuser=%s%s "
                 "session_authorization=%s%s "
                 "DateStyle=%s%s\n"
               , PQSTATNDNULL("server_version")
               , PQSTATNDNULL("client_encoding")
               , PQSTATNDNULL("is_superuser")
               , PQSTATNDNULL("session_authorization")
               , PQSTATNDNULL("DateStyle")
               );
        fprintf( stderr
               , "PQprotocolVersion=%d PQsocket=%d PQbackendPID=%d\n"
               , PQprotocolVersion(pcnn->pgcn)
               , PQsocket(pcnn->pgcn)
               , PQbackendPID(pcnn->pgcn)
               );
        fprintf(stderr, "PQstatus %d\n", PQstatus(pcnn->pgcn));
    }
    (void)PQsetNoticeReceiver(pcnn->pgcn, NoticeReceiver, db);
    (void)PQsetErrorVerbosity(pcnn->pgcn, PQERRORS_VERBOSE);
    (void)PQsetErrorContextVisibility(pcnn->pgcn, PQSHOW_CONTEXT_ALWAYS);
    if  (CONNECTION_BAD == PQstatus(pcnn->pgcn)) {
        ex = EX_NOINPUT;
        errno = ENXIO;
        goto out;
    } else {
        if  (  db->flags & PGOBLIN_NEEDINTIME
            && !(db->flags & (PGOBLIN_INT64TIME | PGOBLIN_FLOATTIME))
            ) {
            if  (!db->intran) {
/*          if  (PQtransactionStatus(pcnn->pgcn) != PQTRANS_INTRANS) { */
                PQEXEC(res, "BEGIN;", PGRES_COMMAND_OK, -1, pcnn->pgcn);
                PQclear(res);
            }
            PQEXEC( res
                  , "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR"
                    " SELECT '2000-01-01 00:00:01'::timestamp"
                    ";"
                    "FETCH 1 FROM pgoblin;"
                  , PGRES_TUPLES_OK
                  , -1
                  , pcnn->pgcn
                  );
            if  (PQgetisnull(res, 0, 0) || (PQgetlength(res, 0, 0) < 8)) {
                ifBLIN_QX0("timestamp");
                ex = EX_DATAERR;
                errno = EINVAL;
            } else if (*((u_char*)(PQgetvalue(res, 0, 0))) == 077) {
                db->flags |= PGOBLIN_FLOATTIME;
                ifBLIN_QX2("float timestamp format");
            } else if (*((u_char*)(PQgetvalue(res, 0, 0))) == 0) {
                db->flags |= PGOBLIN_INT64TIME;
                ifBLIN_QX2("int64 timestamp format");
            } else {
                ifBLIN_QX0("unknown timestamp format");
                ex = EX_DATAERR;
                errno = EFTYPE;
            }
            PQclear(res);
            PQEXEC(res, "CLOSE pgoblin;", PGRES_COMMAND_OK, -1, pcnn->pgcn);
            PQclear(res);
            if  (!db->intran) {
                PQEXEC(res, "END;", PGRES_COMMAND_OK, -1, pcnn->pgcn);
                PQclear(res);
    }   }   }
    if  (!!ex) goto out;
    if  (CONNECTION_BAD == PQstatus(pcnn->pgcn)) {
        ex = EX_NOINPUT;
        errno = ENXIO;
        goto out;
    } else {
        const char *c;
        const char *n;
        int         i;

        if  (!db->intran) {
/*      if  (PQtransactionStatus(pcnn->pgcn) != PQTRANS_INTRANS) { */
            PQEXEC(res, "BEGIN;", PGRES_COMMAND_OK, -1, pcnn->pgcn);
            PQclear(res);
        }
        PQEXEC( res
              , "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR"
                " SELECT oid, typname FROM pg_catalog.pg_type ORDER BY oid;"
                ";"
                "FETCH 9999 FROM pgoblin;"
              , PGRES_TUPLES_OK
              , -1
              , pcnn->pgcn
              );
        if  (!(c = PQcmdStatus(res)) || strncmp("FETCH ", c, 6)) {
            ifBLIN_QX0("PQcmdStatus pg_type");
            ex = EX_DATAERR;
            errno = EINVAL;
        } else if (!(n = PQcmdTuples(res)) || !(0 < (i = strtol(n, NULL, 10)))) {
            ifBLIN_QX0("PQcmdTuples pg_type");
            ex = EX_DATAERR;
            errno = EINVAL;
        } else if (PQntuples(res) != i) {
            ifBLIN_QX0("PQntuples <> %d", i);
            ex = EX_DATAERR;
            errno = EINVAL;
        } else if (!(pcnn->toid = calloc(i, sizeof(u_int32_t)))) {
            ifBLIN_QX0("no MEM");
            ERROUT(EX_OSERR, ENOMEM);
        } else if (!(pcnn->tname = calloc(i, sizeof(char *)))) {
            ifBLIN_QX0("no MEM");
            ERROUT(EX_OSERR, ENOMEM);
        } else {
            for (int k = 0; k < i; ++k) {
                if  (!!PQgetisnull(res, k, 0) || !!PQgetisnull(res, k, 1)) {
                    ifBLIN_QX0("PQgetisnull");
                    ex = EX_DATAERR;
                    errno = EINVAL;
                    break;
                } else if (!(pcnn->tname[k] = strndup( PQgetvalue(res, k, 1)
                                                     , PQgetlength(res, k, 1)
                          ) )                        ) {
                    ex = EX_DATAERR;
                    errno = EINVAL;
                    break;
                } else {
                    pcnn->toid[k] = ntohl(*((u_int32_t *)PQgetvalue(res, k, 0)));
                    pcnn->numoi = k;
        }   }   }
        PQclear(res);
        PQEXEC(res, "CLOSE pgoblin;", PGRES_COMMAND_OK, -1, pcnn->pgcn);
        PQclear(res);
        if  (!db->intran) {
            PQEXEC(res, "END;", PGRES_COMMAND_OK, -1, pcnn->pgcn);
            PQclear(res);
    }   }
out:
    if  (!!ex) ifBLIN_QX0("Connection to PostgreSQL failed: %s", PQerrorMessage(pcnn->pgcn));
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
tarry(pgoblin_conn *db, int timeout) {                              /**
 **                                                                  **
 **********************************************************************/
    pgoblin_db_conn *pcnn = db->conn;
    int              i = 0;
    int              sock;
    struct pollfd    pfd;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    if  ((sock = PQsocket(pcnn->pgcn)) < 0) {
        i = -2;
        errno = EINVAL;
    } else if (timeout != 0) {
        bzero(&pfd, sizeof(pfd));
        pfd.fd = sock;
        pfd.events = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
        i = poll(&pfd, 1, timeout);
    }
    if  (i >= 0) {
        if  (!PQconsumeInput(pcnn->pgcn)) i = -1;
    }
    return(i);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
query(pgoblin_conn *db, pgoblin_mode **rsv, char *qry)    {         /**
 **                                                                  **
 **********************************************************************/
    BLIN_flag        cond = db->flags & PGOBLIN_CR_MASK;
    pgoblin_db_conn *pcnn = db->conn;
    int              doend = 0;
    PGresult        *res = NULL;
    int              ex = 0;
    char            *p;
    char            *q;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    if  (!db) ERROUT(EX_USAGE, EINVAL);
    if  (!qry) {
        ifBLIN_QX1("No query");
        goto out;
    }
    switch (cond) {
    case PGOBLIN_CR_OK:
    case PGOBLIN_CR_TUPL:
    case PGOBLIN_CR_OUT:
    case PGOBLIN_CR_IN:
        ifBLIN_QX4("PQexec %d=%s~", cond, qry);
        if  (!(res = PQexec(pcnn->pgcn, qry))) {
            ifBLIN_QX0("PQ result is absent - fatal error %s", PQerrorMessage(pcnn->pgcn));
            ERROUT(EX_UNAVAILABLE, ENODEV);
        }
        ex = extatus(db->flags, res, "PQexec nom", cond, __func__, __LINE__);
        break;
    case PGOBLIN_CR_SSEL:
        p = "pgoblin";
        goto rest;
    case PGOBLIN_CR_BTUPL:
    case PGOBLIN_CR_BSEL:
        p = "pgoblin BINARY";
    rest:
        MARK_R_CONN_GO(db);;
        ;;  db->flags &= ~PGOBLIN_CR_MASK;
        ;;  db->flags |= PGOBLIN_CR_OK;
        MARK_R_CONN_WENT(db);;
        if  (!db->intran && (ex = query(db, NULL, "BEGIN"))) goto out;
        asprintf(&q, "DECLARE %s NO SCROLL CURSOR FOR %s", p, qry);
        if  (!q) {
            ifBLIN_QX0("No memory #1");
            ERROUT(EX_OSERR, ENOMEM);
        }
        ex = query(db, NULL, q);
        free(q);
        if  (ex) goto out;
        if  (PGOBLIN_CR_BTUPL == cond) {
            doend = 1;
            ifBLIN_QX4("PQexec %d=%s~", cond, "FETCH 2147483647 FROM pgoblin");
            if  (!(res = PQexec(pcnn->pgcn, "FETCH 2147483647 FROM pgoblin"))) {
                ifBLIN_QX0("PQ result is absent - fatal error %s", PQerrorMessage(pcnn->pgcn));
                ERROUT(EX_UNAVAILABLE, ENODEV);
            }
            ex = extatus(db->flags, res, "PQexec bin", cond, __func__, __LINE__);
            if  (!!ex) {
                ifBLIN_QW0("pgoblin_db_query FETCH 2147483647 FROM pgoblin");
                goto out;
            }
            if  (!!(ex = query(db, NULL, "CLOSE pgoblin"))) goto out;
        }
        break;
    default:
        ifBLIN_QX0("illegal cond=%d", cond);
        ex = -1;
    }
out:
    if  (!!doend) ex = query(db, NULL, "END");
    MARK_R_CONN_GO(db);;
    ;;  db->flags &= ~PGOBLIN_CR_MASK;
    ;;  db->flags |= cond;
    MARK_R_CONN_WENT(db);;
    if  (!rsv) {
        if  (!!res) PQclear(res); /* XXXX */
    } else if (!!res) {
        if  (!(*rsv = resv(db))) {
            ifBLIN_QX0("No memory #2");
            ex = EX_OSERR;
            errno = ENOMEM;
        } else {
            (*rsv)->res = res;
        }
    } else {
        *rsv = NULL;
    }
    return(ex);
#   undef blin_internal_flags
}

#ifdef PG_DIAG_INTERNAL_POSITION

static int
/**********************************************************************
 **                                                                  **/
prepare(pgoblin_conn *db, pgoblin_mode **rsv, char *qry) {          /**
 **                                                                  **
 **********************************************************************/
    pgoblin_db_conn *pcnn = db->conn;
    int              ex = 0;
    PGresult        *res;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    res = NULL;
    if  (!db) ERROUT(EX_USAGE, EINVAL);
    ifBLIN_QX4("PQprepare =%s~", qry);
    if  (!(res = PQprepare(pcnn->pgcn, "", qry, 0, NULL))) {
        ifBLIN_QX0("PQ result is absent - fatal error %s", PQerrorMessage(pcnn->pgcn));
        ERROUT(EX_UNAVAILABLE, ENODEV);
    }
    ex = extatus(db->flags, res, "PQprepare", PGOBLIN_CR_OK, __func__, __LINE__);
out:
    if  (!rsv) {
        PQclear(res); /* XXXX */
    } else if (!!res) {
        if  (!(*rsv = resv(db))) {
            ifBLIN_QX0("No memory #2");
            ex = EX_OSERR;
            errno = ENOMEM;
        } else {
            (*rsv)->res = res;
        }
    } else {
        *rsv = NULL;
    }
    return(ex);
#   undef blin_internal_flags
}

#endif

static int
/*************************************************************************************
 **                                                                                 **/
execute(pgoblin_conn *db, pgoblin_mode **rsv, const void *qp, pgoblin_mode *qin) { /**
 **                                                                                 **
 *************************************************************************************/
    pgoblin_db_conn *pcnn = db->conn;
    BLIN_flag        cond = db->flags & PGOBLIN_CR_MASK;
    char const      *kind;
    int              imdx;
    const char     **val  = NULL;
    PGresult        *res  = NULL;
    const char      *qry  = "";
    int             *len  = NULL;
    int              col;
    int              row;
    int              sz;
    int              ex   = 0;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    res = NULL;
    if  (!db) {
        ERROUT(EX_USAGE, EINVAL);
    }
    if  ((PGOBLIN_CR_OK != cond) && (PGOBLIN_CR_TUPL != cond)) {
        ifBLIN_QX0("execute pgsql illegal cond=%d", cond);
        ERROUT(EX_USAGE, EINVAL);
    }
    col = pgoblin_db_resinfo(PGOBLIN_Nfields, qin);
    row = pgoblin_db_resinfo(PGOBLIN_Ntuples, qin);
    if  (db->flags & PGOBLIN_ALTERQ) {
        kind = "PQexecPrepared";
        if  (!(res = PQdescribePrepared(pcnn->pgcn, qry))) {
            ifBLIN_QX0("PQdescribePrepared is absent - fatal error %s", PQerrorMessage(pcnn->pgcn));
            ERROUT(EX_UNAVAILABLE, ENODEV);
        }
        imdx = PQnparams(res);
        PQclear(res);
        res = NULL;
    } else if (!qp) {
        ifBLIN_QX0("NULL qp");
        ERROUT(EX_USAGE, EINVAL);
    } else {
        kind = "PQexecParams";
        qry = qp;
        imdx = col;
    }
    if  (col < imdx) sz = imdx; else sz = col;
    ifBLIN_QX4("%d rows x %d cols (%d parms) %d", row, col, imdx, sz);
    if  (!(val = malloc(sizeof(char *) * sz))) {
        ifBLIN_QW0("pgoblin_select #1: No mem");
        ex = EX_OSERR;
        goto out;
    }
    if  (!(len = malloc(sizeof(int) * sz))) {
        ifBLIN_QW0("pgoblin_select #2: No mem");
        free(val);
        val = NULL;
        ex = EX_OSERR;
        goto out;
    }
    for (int i = 0; i < row; i++) {
        for (int j = 0; j < col; j++) {
            val[j] = pgoblin_db_getvalue(qin, i, j);
            len[j] = pgoblin_db_valinfo(PGOBLIN_Length, qin, i, j);
        }
        for (int j = col; j < imdx; j++) {
            val[j] = NULL;
            len[j] = -1;
        }
        ifBLIN_QX4("%s %d=%s~", kind, cond, qry);
        if  (!(res = (db->flags & PGOBLIN_ALTERQ)
                   ? PQexecPrepared(pcnn->pgcn, qry, imdx,       val, len, NULL, 0)
                   : PQexecParams(  pcnn->pgcn, qry, imdx, NULL, val, len, NULL, 0)
            ) ) {
            ifBLIN_QX0("PQ result is absent - fatal error %s", PQerrorMessage(pcnn->pgcn));
            ERROUT(EX_UNAVAILABLE, ENODEV);
        }
        if  (!!(ex = extatus(db->flags, res, kind, cond, __func__, __LINE__))) break;
        if  (i + 1 < row) {
            PQclear(res);
            res = NULL;
    }   }
out:
    if  (val) free(val);
    if  (len) free(len);
    if  (!!rsv) *rsv = NULL;
    if  (!!res) {
        if  (!rsv) {
            PQclear(res); /* XXXX */
            res = NULL;
        } else if (!(*rsv = resv(db))) {
            ifBLIN_QX0("query: No memory #2");
            ex = EX_OSERR;
            errno = ENOMEM;
        } else {
            (*rsv)->res = res;
    }   }
    return(ex);
#   undef blin_internal_flags
}

static void *
/**********************************************************************
 **                                                                  **/
notifies(pgoblin_conn *db) {                                        /**
 **                                                                  **
 **********************************************************************/
    pgoblin_db_conn    *pcnn = db->conn;
    pgoblin_longnotify *c    = NULL;
    PGnotify           *n;
    char               *o    = NULL;

    n = PQnotifies(pcnn->pgcn);
    if  (!!n) {
        errno = 0;
        c = calloc(1, sizeof(pgoblin_longnotify) + strlen(n->relname) + strlen(n->extra) + 3);
        if  (!c) {
            ifBLIN_QW0("For pgsql notifies");
            goto out;
        }
        c->majorlen = strlen(n->relname) + 1;
        c->minorlen = strlen(n->extra) + 1;
        c->ident = n->be_pid;
        o = stpncpy(c->tail, n->relname, c->majorlen);
        strncpy(++o, n->extra, c->minorlen);
        PQfreemem(n);
    }
out:
    return(c);
}

static int64_t
/**********************************************************************
 **                                                                  **/
resinfo(int kind, pgoblin_mode *rsv) {                              /**
 **                                                                  **
 **********************************************************************/
    int   ex;
    char *c;

    switch (kind) {
    case PGOBLIN_Ntuples: ex =   PQntuples(rsv->res); break;
    case PGOBLIN_Nfields: ex =   PQnfields(rsv->res); break;
    case PGOBLIN_Nchange:
        c = PQcmdTuples(rsv->res);
        ex = *c ? strtoul(c, NULL, 10) : -1;
        break;
    case PGOBLIN_LastOid: ex =  PQoidValue(rsv->res); break;
    default             : ex = -1;
    }
    return(ex);
}

static char *
/**********************************************************************
 **                                                                  **/
erinfo(pgoblin_mode *rsv) {                                         /**
 **                                                                  **
 **********************************************************************/
    char *cx = NULL;

#   define blin_internal_flags (rsv->flags & BLIN_MASK)
    if  (!!rsv->mess) goto out;
    if  (!rsv->res) {
        if  (asprintf(&cx, "No results") < 0) warn("subinfo(ResMess) asprintf");
    } else ifBLIN_QO4 {
        if  ( asprintf( &cx
#ifdef PG_DIAG_INTERNAL_POSITION
                      , "%s %s %s\n%s\n%s\n%s\n%s\n%s\nfunction %s at %s line of %s file\n%s"
#else
                      , "%s %s %s\n%s\n%s\n%s\nfunction %s at %s line of %s file\n%s"
#endif
                      , PQresultErrorField(rsv->res, PG_DIAG_SEVERITY)
                      , PQresultErrorField(rsv->res, PG_DIAG_SQLSTATE)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_PRIMARY)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_DETAIL)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_HINT)
                      , PQresultErrorField(rsv->res, PG_DIAG_STATEMENT_POSITION)
#ifdef PG_DIAG_INTERNAL_POSITION
                      , PQresultErrorField(rsv->res, PG_DIAG_INTERNAL_QUERY)
                      , PQresultErrorField(rsv->res, PG_DIAG_INTERNAL_POSITION)
#endif
                      , PQresultErrorField(rsv->res, PG_DIAG_SOURCE_FUNCTION)
                      , PQresultErrorField(rsv->res, PG_DIAG_SOURCE_LINE)
                      , PQresultErrorField(rsv->res, PG_DIAG_SOURCE_FILE)
                      , PQresultErrorField(rsv->res, PG_DIAG_CONTEXT)
                      )
              < 0
            ) warn("subinfo(ResMess) asprintf");
    } else ifBLIN_QO3 {
        if  ( asprintf( &cx
#ifdef PG_DIAG_INTERNAL_POSITION
                      , "%s %s %s\n%s\n%s\n%s\n%s\n%s\nfunction %s at %s line of %s file"
#else
                      , "%s %s %s\n%s\n%s\n%s\nfunction %s at %s line of %s file"
#endif
                      , PQresultErrorField(rsv->res, PG_DIAG_SEVERITY)
                      , PQresultErrorField(rsv->res, PG_DIAG_SQLSTATE)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_PRIMARY)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_DETAIL)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_HINT)
                      , PQresultErrorField(rsv->res, PG_DIAG_STATEMENT_POSITION)
#ifdef PG_DIAG_INTERNAL_POSITION
                      , PQresultErrorField(rsv->res, PG_DIAG_INTERNAL_QUERY)
                      , PQresultErrorField(rsv->res, PG_DIAG_INTERNAL_POSITION)
#endif
                      , PQresultErrorField(rsv->res, PG_DIAG_SOURCE_FUNCTION)
                      , PQresultErrorField(rsv->res, PG_DIAG_SOURCE_LINE)
                      , PQresultErrorField(rsv->res, PG_DIAG_SOURCE_FILE)
                      )
              < 0
            ) warn("subinfo(ResMess) asprintf");
    } else ifBLIN_QO2 {
        if  ( asprintf( &cx
#ifdef PG_DIAG_INTERNAL_POSITION
                      , "%s %s %s\n%s\n%s\n%s\n%s\n%s"
#else
                      , "%s %s %s\n%s\n%s\n%s"
#endif
                      , PQresultErrorField(rsv->res, PG_DIAG_SEVERITY)
                      , PQresultErrorField(rsv->res, PG_DIAG_SQLSTATE)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_PRIMARY)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_DETAIL)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_HINT)
                      , PQresultErrorField(rsv->res, PG_DIAG_STATEMENT_POSITION)
#ifdef PG_DIAG_INTERNAL_POSITION
                      , PQresultErrorField(rsv->res, PG_DIAG_INTERNAL_QUERY)
                      , PQresultErrorField(rsv->res, PG_DIAG_INTERNAL_POSITION)
#endif
                      )
              < 0
            ) warn("subinfo(ResMess) asprintf");
    } else ifBLIN_QO1 {
        if  ( asprintf( &cx
                      , "%s %s %s\n%s\n%s\n%s"
                      , PQresultErrorField(rsv->res, PG_DIAG_SEVERITY)
                      , PQresultErrorField(rsv->res, PG_DIAG_SQLSTATE)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_PRIMARY)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_DETAIL)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_HINT)
                      , PQresultErrorField(rsv->res, PG_DIAG_STATEMENT_POSITION)
                      )
              < 0
            ) warn("subinfo(ResMess) asprintf");
    } else ifBLIN_QO0 {
        if  ( asprintf( &cx
                      , "%s %s %s\n%s"
                      , PQresultErrorField(rsv->res, PG_DIAG_SEVERITY)
                      , PQresultErrorField(rsv->res, PG_DIAG_SQLSTATE)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_PRIMARY)
                      , PQresultErrorField(rsv->res, PG_DIAG_MESSAGE_DETAIL)
                      )
              < 0
            ) warn("subinfo(ResMess) asprintf");
    }
    rsv->mess = cx;
out:
    return(rsv->mess);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
valinfo(int kind, pgoblin_mode *rsv, int row, int col) {            /**
 **                                                                  **
 **********************************************************************/
    int ex;

    switch (kind) {
    case PGOBLIN_IsNull : ex = PQgetisnull(rsv->res, row, col); break;
    case PGOBLIN_Length : ex = PQgetlength(rsv->res, row, col); break;
    case PGOBLIN_BinForm: ex =   PQfformat(rsv->res,      col); break;
    case PGOBLIN_TypeOid: ex =     PQftype(rsv->res,      col); break;
    case PGOBLIN_TypeMod: ex =      PQfmod(rsv->res,      col); break;
    case PGOBLIN_ColSize: ex =     PQfsize(rsv->res,      col); break;
    default             :
        ex = INT_MIN;
        errno = EDOOFUS;
    }
    return(ex);
}

static const char *
/**********************************************************************
 **                                                                  **/
subinfo(int kind, pgoblin_mode *rsv, int sub) {                     /**
 **                                                                  **
 **********************************************************************/
    const char *cx = NULL;
    u_int32_t   ex;

#   define blin_internal_flags (rsv->flags & BLIN_MASK)
    switch (kind) {
    case PGOBLIN_ColName:
        cx = PQfname(rsv->res, sub);
        break;
    case PGOBLIN_CmdStat:
        cx = PQcmdStatus(rsv->res);
        break;
    case PGOBLIN_TypeNm :
        ex = valinfo(PGOBLIN_TypeOid, rsv, 0, sub);
        for ( int l = 0, r = rsv->pcnn->numoi, m; l < r;) {
            m = (l + r) / 2;
            if  (rsv->pcnn->toid[m] == ex) {
                cx = rsv->pcnn->tname[m];
                break;
            } else if (rsv->pcnn->toid[m] < ex) {
                l = m;
            } else {
                r = m;
        }   }
        break;
    default             :
        cx = NULL;
    }
    return(cx);
#   undef blin_internal_flags
}

static char *
/**********************************************************************
 **                                                                  **/
getvalue(pgoblin_mode *rsv, int row, int col) {                     /**
 **                                                                  **
 **********************************************************************/
    char *c = NULL;

    if  (!valinfo(PGOBLIN_IsNull, rsv, row, col)) {
        c = PQgetvalue(rsv->res, row, col);
    }
    return(c);
}

static int
/**********************************************************************
 **                                                                  **/
finish(u_int32_t flags, int kind, void **entity) {                  /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (flags & BLIN_MASK)
    int ex = EX_OK;

    switch (kind) {
    case PGOBLIN_Connect: {
            pgoblin_db_conn *pcnn;

            if  (entity && *entity) {
                pcnn = *entity;
                PQfinish(pcnn->pgcn);
                if  (pcnn->tname) for (int i = 0; i < pcnn->numoi; ++i) free(pcnn->tname[i]);
                free(pcnn->toid);
                *entity = NULL;
        }   }
        break;
    case PGOBLIN_FrCOPY:
    case PGOBLIN_FrNote:
        if  (entity && *entity) {
            free(*entity);
            *entity = NULL;
        }
        break;
    default:
        ex = EX_UNAVAILABLE;
        errno = ENOTSUP;
    }
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
clear(pgoblin_mode **res) {                                         /**
 **                                                                  **
 **********************************************************************/
    int ex = EX_OK;

    if  (!!res && !!*res) {
        if  ((*res)->res) PQclear((*res)->res);
        if  ((*res)->mess) free((*res)->mess);
        free(*res);
        *res = NULL;
    }
    return(ex);
}

static int
/****************************************************************************
 **                                                                        **/
getstream(pgoblin_conn *db, pgoblin_mode **res, int quant) { /**
 **                                                                        **
 ****************************************************************************/
    BLIN_flag        cond = db->flags & PGOBLIN_CR_MASK;
    int              ex = EX_OK;
    char            *qry;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    if  (!res) {
        ifBLIN_QX0("no res");
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!!*res) clear(res);
    if  (!db) ERROUT(EX_USAGE, EINVAL);
    if  ((cond != PGOBLIN_CR_SSEL) && (cond != PGOBLIN_CR_BSEL)) {
        ifBLIN_QX0("Illegal cond %d", cond);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (asprintf(&qry, "FETCH %d FROM pgoblin", quant) < 0) {
        ifBLIN_QX0("No memory");
        ERROUT(EX_OSERR, ENOMEM);
    }
    MARK_R_CONN_GO(db);;
    ;;  db->flags &= ~PGOBLIN_CR_MASK;
    ;;  db->flags |= PGOBLIN_CR_TUPL;
    MARK_R_CONN_WENT(db);;
    ex = query(db, res, qry);
    free(qry);
out:
    MARK_R_CONN_GO(db);;
    ;;  db->flags &= ~PGOBLIN_CR_MASK;
    ;;  db->flags |= cond;
    MARK_R_CONN_WENT(db);;
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
getcopy(pgoblin_conn *db, void **buf) {                             /**
 **                                                                  **
 **********************************************************************/
    pgoblin_db_conn *pcnn = db->conn;
    int              ex;

    ex = PQgetCopyData(pcnn->pgcn, (char**)buf, 0);
    return((ex < 0) ? (ex + 1) : ex);
}

static int
/**********************************************************************
 **                                                                  **/
putcopy(pgoblin_conn *db, const void *buf, int nby) {               /**
 **                                                                  **
 **********************************************************************/
    pgoblin_db_conn *pcnn = db->conn;

    return(PQputCopyData(pcnn->pgcn, buf, nby));
}

static int
/**********************************************************************
 **                                                                  **/
endstream(pgoblin_conn *db) {                                       /**
 **                                                                  **
 **********************************************************************/
    pgoblin_db_conn *pcnn = db->conn;
    int              ex   = EX_OK;
    BLIN_flag        cond;
    PGresult        *res;
    ExecStatusType   st;

#   define blin_internal_flags (db->flags & BLIN_MASK)
    cond = db->flags & PGOBLIN_CR_MASK;
    switch (cond) {
    case PGOBLIN_CR_IN:
        if  (PQputCopyEnd(pcnn->pgcn, NULL) < 1) {
            ifBLIN_QX0("endcopy failed: %s", PQerrorMessage(pcnn->pgcn));
            ERROUT(-EX_DATAERR, EINVAL);
        }
        /* FALLTHRU */
    case PGOBLIN_CR_OUT:
        while ((res = PQgetResult(pcnn->pgcn))) {
            st = PQresultStatus(res);
            ifBLIN_QX4("PQresultStatus %d %s", st, PQresStatus(st));
            if  (st != PGRES_COMMAND_OK) {
                ifBLIN_QX0("%s", PQresultErrorMessage(res));
                ex = EX_SOFTWARE;
                break;
            }
            PQclear(res);
            res = NULL;
        }
        break;
    case PGOBLIN_CR_SSEL:
    case PGOBLIN_CR_BSEL:
        MARK_R_CONN_GO(db);;
        ;;  db->flags &= ~PGOBLIN_CR_MASK;
        ;;  db->flags |= PGOBLIN_CR_OK;
        MARK_R_CONN_WENT(db);;
        ex = query(db, NULL, "CLOSE pgoblin");
        if  (!db->intran) {
            ex |= query(db, NULL, "END");
        }
        break;
    default:
        ifBLIN_QX0("- %d", cond);
        ex = EX_USAGE;
    }
out:
    MARK_R_CONN_GO(db);;
    ;;  db->flags &= ~PGOBLIN_CR_MASK;
    ;;  db->flags |= cond;
    MARK_R_CONN_WENT(db);;
    if  (ex) errno = EINVAL;
    return(ex);
#   undef blin_internal_flags
}

pgoblin_dbases PGOBLIN_MODULE =
{ ( PGOBLIN_DB_COPYIN
  | PGOBLIN_DB_COPYOUT
#ifdef PG_DIAG_INTERNAL_POSITION
  | PGOBLIN_DB_PREPARE
#endif
  | PGOBLIN_LONGNOTIFY
  )
, "3pGoblin-" VERS
, "pgsql\0" VERS
, 0
, init
, fini
, shurecon
, query
#ifdef PG_DIAG_INTERNAL_POSITION
, prepare
#else
, NULL
#endif
, execute
, resinfo
, subinfo
, erinfo
, valinfo
, getvalue
, getstream
, getcopy
, putcopy
, endstream
, tarry
, notifies
, finish
, clear
, NULL
, NULL
};
